"""token_counter contains TokenCounter."""
import tiktoken

from echo_artistry import utils


class TokenCounter:
    """A class for counting tokens."""

    def __init__(self, model_name=utils.DEFAULT_MODEL_NAME):
        self.encoding = tiktoken.encoding_for_model(model_name)
        self.model_token_length = utils.MODEL_TOKEN_LENGTH_MAPPING[model_name][
            "token_length"
        ]

    def count_tokens(self, text):
        """Count the number of tokens in a text.

        Args:
            text (str): The text to count the tokens of.

        Returns:
            int: The number of tokens in the text.
        """
        token_count = len(self.encoding.encode(text))
        return token_count



"""comic_story_generator contains ComicStoryGenerator and CompositeOption class."""
import os
from enum import Enum

from echo_artistry import utils

MAX_RETRIES = 5
MAX_FILE_NAME_LENGTH = 15


class CompositeOption(Enum):
    """CompositeOption defines the composite option for the comic."""

    SINGLE_IMAGE = "single_image"


class ComicStoryGenerator:
    """ComicStoryGenerator generates comic stories."""

    def __init__(
        self,
        client,
        output_path="comic_stories",
        max_retries=MAX_RETRIES,
        store_comic_story=True,
    ):
        self.client = client
        self.max_retries = max_retries
        self.output_path = output_path
        self.store_comic_stories = store_comic_story
        if self.store_comic_stories:
            if not os.path.exists(self.output_path):
                os.makedirs(self.output_path)

    def rewrite_text_to_comic_story(self, text):
        """Rewrite a text to a comic story.

        Args:
            text (str): The text to rewrite to a comic story.

        Returns:
            str: The comic story.
        """
        msg = [
            {
                "role": "system",
                "content": f"Rewrite the given text to a comic.\n"
                f"At the beginning name the amount of panels, like: "
                           f"Comic strip with [NAME_NUMBER] panels.\n"
                f"Than describe each panel with a few and short sentence.",
            },
            {"role": "user", "content": f"TEXT: {text}"},
        ]
        for _ in range(MAX_RETRIES):
            comic_story = self.client.generate_answer(msg)
            if (
                len(comic_story)
                < utils.IMAGE_CHARACTER_LENGTH_MAPPING[self.client.image_model_name][
                    "character_length"
                ]
            ):
                return comic_story
            else:
                msg.append({"role": "assistant", "content": comic_story})
                msg.append(
                    {"role": "user", "content": f"Keep the content shorter shorter."},
                )

        raise Exception("Max retries exceeded.")

    def generate_story(
        self,
        text,
        composite_option: CompositeOption = CompositeOption.SINGLE_IMAGE,
        file_name="comic.txt",
    ):
        """Generate a comic description from a text.

        Args:
            text (str): The text to generate the comic description from.
            composite_option (CompositeOption): The composite option for the comic.
            file_name (str): The name of the file to store the comic story.

        Returns:
            str: The generated description.
        """
        if composite_option == CompositeOption.SINGLE_IMAGE:
            comic_story = self.rewrite_text_to_comic_story(text)
        else:
            raise NotImplementedError(
                f"Composite option {composite_option} not implemented.",
            )

        if self.store_comic_stories:
            comic_story_path = f"{self.output_path}/{file_name}"
            utils.write_text_to_file(comic_story, comic_story_path)

        return comic_story



"""comic_image_generator contains the ComicImageGenerator class."""

import os

MAX_RETRIES = 5


class ComicImageGenerator:
    """A class for generating a comic image from a text."""

    def __init__(
        self,
        client,
        output_path="comic_images",
        store_comic_images=True,
    ):
        self.client = client
        self.output_path = output_path
        self.store_comic_images = store_comic_images
        if self.store_comic_images:
            if not os.path.exists(self.output_path):
                os.makedirs(self.output_path)

    def generate_image(self, text, file_name="comic.png"):
        """Generate a comic description from a text.

        Args:
            text (str): The text to generate the
            composite_option (CompositeOption): The composite option for the comic.
            file_name (str): The name of the file to store the comic description.

        Returns:
            str: The generated description.
        """
        comic_image = self.client.generate_image(text)

        if self.store_comic_images:
            comic_image_path = os.path.join(self.output_path, file_name)
            comic_image.save(comic_image_path)

        return comic_image



"""transcriber contains Transcriber class."""

import os
from pathlib import Path

import whisper

from echo_artistry import utils


class Transcriber:
    """Extracts audio from a video file and transcribes it to text."""

    def __init__(self, output_path="audios", store_transcription=True):
        self.transcribe_model = whisper.load_model("base")
        self.store_transcription = store_transcription
        self.output_path = output_path
        if self.store_transcription:
            if not os.path.exists(self.output_path):
                os.makedirs(self.output_path)

    def _storing_transcription(self, transcript_result, audio_file_path):
        audio_file_name = Path(audio_file_path).stem
        transcription_file_path = Path(
            self.output_path,
            f"transcription_{audio_file_name}.txt",
        )
        utils.write_text_to_file(transcript_result, str(transcription_file_path))

    def transcribe_audio(self, audio_file_path):
        """Transcribes audio to text.

        Args:
            audio_file_path (str): The path to the audio file.

        Returns:
            str: The transcription of the audio file.
        """
        transcript_result = self.transcribe_model.transcribe(audio_file_path)["text"]

        if self.store_transcription:
            self._storing_transcription(transcript_result, audio_file_path)
        return transcript_result



"""openai_exceptions contain OpenAIError and ContentPolicyViolation class."""

class OpenAIError(Exception):
    """General error in OpenAI Client."""

    pass


class ContentPolicyViolation(OpenAIError):
    """Content policy violation."""

    pass



"""openai_client contains OpenAIClient class."""
from io import BytesIO

import requests
from openai import OpenAI
from PIL import Image

from echo_artistry import exceptions
from echo_artistry.utils import helper


class OpenAIClient:
    """OpenAI API client."""

    def __init__(
        self,
        cost_manager,
        model_name=helper.DEFAULT_MODEL_NAME,
        image_model_name=helper.DEFAULT_IMAGE_MODEL_NAME,
        image_size=helper.DEFAULT_IMAGE_SIZE,
        image_quality=helper.DEFAULT_IMAGE_QUALITY,
    ):
        self.cost_manager = cost_manager
        self.model_name = model_name
        self.image_model_name = image_model_name
        self.image_size = image_size
        self.image_quality = image_quality
        self.client = OpenAI()

    def _calculate_cost(self, messages=None, image=None, is_input=True):
        if image:
            self.cost_manager.calculate_cost_image(image)
        if messages:
            self.cost_manager.calculate_cost_messages(messages, is_input=is_input)

    def generate_answer(self, messages):
        """Generate answer.

        Args:
            messages (list(str)): List of messages.

        Returns:
            str: Generated answer.
        """
        self._calculate_cost(messages=messages, is_input=True)
        try:
            response = self.client.chat.completions.create(
                model=self.model_name,
                messages=messages,
            )
            helper.logger.debug(
                f"Messages: {messages} Response: {response.choices[0].message.content}",
            )
            result_text = response.choices[0].message.content
        except Exception as e:
            raise exceptions.OpenAIError(f"An unexpected error occurred: {e}") from e

        result_message = [{"role": "assistant", "content": result_text}]
        self._calculate_cost(messages=result_message, is_input=False)
        return result_text

    def _retrieve_image_from_url(self, url):
        response = requests.get(url)
        return Image.open(BytesIO(response.content))

    def generate_image(self, prompt):
        """Generate image.

        Args:
            prompt (str): Prompt.

        Returns:
            str: Generated image.
        """
        try:
            response = self.client.images.generate(
                model=self.image_model_name,
                prompt=prompt,
                size=self.image_size,
                quality="standard",
                n=1,
            )
        except Exception as e:
            if "content_policy_violation" in str(e):
                raise exceptions.ContentPolicyViolation(
                    "The given text contains content which "
                    "violates the OpenAI content policy.",
                )
            else:
                raise exceptions.OpenAIError(f"An unexpected error occurred: {e}") from e
        image = self._retrieve_image_from_url(response.data[0].url)
        self._calculate_cost(image=image, is_input=True)
        return image



"""helper module provides utility functions and constants."""

import logging

import tiktoken

logging.basicConfig(
    filename="echo_artistry.log",
    level=logging.DEBUG,
    format="%(asctime)s - %(levelname)s - %(message)s",
)
logger = logging.getLogger()

MODEL_TOKEN_LENGTH_MAPPING = {
    "gpt-3.5-turbo-1106": {
        "token_length": 16385,
        "input_token_cost": 0.0010,
        "output_token_cost": 0.0020,
    },
    "gpt-4-1106-preview": {
        "token_length": 128000,
        "input_token_cost": 0.01,
        "output_token_cost": 0.03,
    },
    "gpt-4": {
        "token_length": 8192,
        "input_token_cost": 0.03,
        "output_token_cost": 0.06,
    },
    "gpt-4-32k": {
        "token_length": 32768,
        "input_token_cost": 0.06,
        "output_token_cost": 0.12,
    },
}

IMAGE_CHARACTER_LENGTH_MAPPING = {
    "dall-e-2": {
        "character_length": 1000,
        "cost_per_image": {
            "standard": {
                "1024x1024": 0.020,
                "512x512": 0.018,
                "256x256": 0.016,
            },
        },
    },
    "dall-e-3": {
        "character_length": 4000,
        "cost_per_image": {
            "standard": {
                "1024x1024": 0.040,
                "1024x1792": 0.080,
                "1792x1024": 0.080,
            },
            "hd": {
                "1024x1024": 0.080,
                "1024x1792": 0.120,
                "1792x1024": 0.120,
            },
        },
    },
}

DEFAULT_MODEL_NAME = "gpt-3.5-turbo-1106"
DEFAULT_IMAGE_MODEL_NAME = "dall-e-3"
DEFAULT_IMAGE_SIZE = "1792x1024"
DEFAULT_IMAGE_QUALITY = "standard"
MAX_RETRIES = 5


def write_text_to_file(text, file_path):
    """Write text to a file.

    Args:
        text (str): The text to write.
        file_path (str): The path to the file.
    """
    with open(file_path, "w") as file:
        file.write(text)


def get_text_from_file(file_path):
    """Get the text from a file.

    Args:
        file_path (str): The path to the file.

    Returns:
        str: The text from the file.
    """
    with open(file_path, "r") as file:
        text = file.read()
    return text



# noqa: D104

from .helper import (
    DEFAULT_IMAGE_MODEL_NAME,
    DEFAULT_IMAGE_QUALITY,
    DEFAULT_IMAGE_SIZE,
    DEFAULT_MODEL_NAME,
    IMAGE_CHARACTER_LENGTH_MAPPING,
    MAX_RETRIES,
    MODEL_TOKEN_LENGTH_MAPPING,
    get_text_from_file,
    logger,
    write_text_to_file,
)
from .openai_client import OpenAIClient



"""Generate a blog post from a YouTube video."""

import argparse
import os

from tqdm import tqdm

from echo_artistry.generators import (
    ComicImageGenerator,
    ComicStoryGenerator,
    FileNameGenerator,
)
from echo_artistry.services import CostManager, TokenCounter, Transcriber
from echo_artistry.utils import OpenAIClient, helper


def args_call():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Generate stunning art stories based on your voice message.",
    )
    parser.add_argument(
        "audio_path",
        type=str,
        help="The path to the audio file.",
    )
    parser.add_argument(
        "output_path",
        type=str,
        help="The directory to save the summary file.",
    )
    parser.add_argument(
        "api_key",
        type=str,
        help="The API key for openai API.",
    )
    parser.add_argument(
        "--model_name",
        type=str,
        default="gpt-3.5-turbo-1106",
        help="The model name used as blog generator.",
    )
    args = parser.parse_args()
    main(args.audio_path, args.output_path, args.api_key, args.model_name)


def main(audio_path, output_path, api_key, model_name):
    """Download, transcribe, and generate comics from a voice memo.

    Args:
    ----
        audio_path (str): The path to the audio file.
        output_path (str): The directory to save the summary file.
        api_key (str): The API key for openai API.
        model_name (str): The model name used as blog generator.
    """
    os.environ["OPENAI_API_KEY"] = api_key
    os.environ["TOKENIZERS_PARALLELISM"] = "false"
    token_counter = TokenCounter(model_name=model_name)
    cost_manager = CostManager(token_counter, model_name=model_name)
    client = OpenAIClient(cost_manager, model_name=model_name)
    transcriber = Transcriber(output_path=output_path)
    file_name_generator = FileNameGenerator(client)
    comic_story_generator = ComicStoryGenerator(client, output_path=output_path)
    comic_image_generator = ComicImageGenerator(client, output_path=output_path)

    tasks = [
        "Transcribing audio",
        "Generate file name",
        "Generating comic story",
        "Generating comic",
    ]

    with tqdm(total=len(tasks)) as pbar:
        transcription = transcriber.transcribe_audio(audio_path)
        pbar.update(1)
        file_name = file_name_generator.generate_file_name(transcription)
        pbar.update(1)
        comic_file_name = f"{file_name}.txt"
        comic_story = comic_story_generator.generate_story(
            transcription,
            file_name=comic_file_name,
        )
        pbar.update(1)
        comic_image_file_name = f"{file_name}.png"
        comic_image_generator.generate_image(comic_story, file_name=comic_image_file_name)
        pbar.update(1)

        total_cost = cost_manager.get_total_cost()
        helper.logging.info(f"Comic is generated.\nTotal cost: {total_cost} USD")


if __name__ == "__main__":
    args_call()



"""cost_management contains CostManager."""

from echo_artistry.utils import helper


class CostManager:
    """CostManager is used to calculate the cost of a text."""

    total_cost = 0

    def __init__(
        self,
        token_counter,
        model_name=helper.DEFAULT_MODEL_NAME,
        image_model_name=helper.DEFAULT_IMAGE_MODEL_NAME,
        image_quality=helper.DEFAULT_IMAGE_QUALITY,
        image_size=helper.DEFAULT_IMAGE_SIZE,
    ):
        self.token_counter = token_counter
        self.model_name = model_name
        self.image_model_name = image_model_name
        self.image_quality = image_quality
        self.image_size = image_size
        self.input_token_cost = helper.MODEL_TOKEN_LENGTH_MAPPING[model_name][
            "input_token_cost"
        ]
        self.output_token_cost = helper.MODEL_TOKEN_LENGTH_MAPPING[model_name][
            "output_token_cost"
        ]
        self.per_n_tokens = 1000

    def calculate_cost_token(self, token_count, is_input=True):
        """Calculate the cost of a text.

        Args:
            token_count (int): The number of tokens in the text.
            is_input (bool, optional): Whether text is input or output. Defaults to True.

        Returns:
            float: The cost of the text.
        """
        if is_input:
            cost = (token_count / self.per_n_tokens) * self.input_token_cost
        else:
            cost = (token_count / self.per_n_tokens) * self.output_token_cost
        self.total_cost += cost
        return cost

    def calculate_cost_text(self, text, is_input=True):
        """Calculate the cost of a text.

        Args:
            text (str): The text to calculate the cost of.
            is_input (bool, optional): Whether the text is an input or an output.
                Defaults to True.

        Returns:
            float: The cost of the text.
        """
        token_count = self.token_counter.count_tokens(text)
        return self.calculate_cost_token(token_count, is_input=is_input)

    def calculate_cost_messages(self, messages, is_input=True):
        """Calculate the cost of a message.

        Args:
            messages (list): The message to calculate the cost of.
            is_input (bool, optional): Whether the message is an input or an output.
                Defaults to True.

        Returns:
            float: The cost of the message.
        """
        for message in messages:
            text = message["content"]
            self.calculate_cost_text(text, is_input=is_input)

    def calculate_cost_image(self, image):
        """Calculate the cost of an image.

        Args:
            image (Image): The image to calculate the cost of.

        Returns:
            float: The cost of the image.
        """
        image_size = image.size
        image_size_string = f"{image_size[0]}x{image_size[1]}"
        cost = helper.IMAGE_CHARACTER_LENGTH_MAPPING[self.image_model_name][
            "cost_per_image"
        ][self.image_quality][image_size_string]
        self.total_cost += cost
        return cost

    def get_total_cost(self):
        """Get the total cost of the text.

        Returns:
            float: The total cost of the text.
        """
        return self.total_cost



"""file_name_generator contains FileNameGenerator."""

MAX_RETRIES = 5
MAX_FILE_NAME_LENGTH = 35


class FileNameGenerator:
    """A class for generating a file name from a text."""

    def __init__(
        self,
        client,
        max_retries=MAX_RETRIES,
        max_file_name_length=MAX_FILE_NAME_LENGTH,
    ):
        self.client = client
        self.max_file_name_length = max_file_name_length
        self.max_retries = max_retries

    @staticmethod
    def _validate_file_name(filename):
        if len(filename) < MAX_FILE_NAME_LENGTH:
            return True
        else:
            return False

    @staticmethod
    def _sanitize_string(topic):
        topic = topic.lower()
        sanitized = topic.replace(" ", "_")
        sanitized = "".join(char for char in sanitized if char.isalnum() or char == "_")

        return sanitized

    def generate_file_name(self, text):
        """Generate a file name from a text.

        Args:
            text (str): The text to generate a file name from.

        Returns:
            str: The generated file name.
        """
        msg = [
            {
                "role": "system",
                "content": f"Name one important topic in following text.",
            },
            {"role": "user", "content": f"TEXT: {text} \n One topic: [TOPIC]"},
        ]
        for _ in range(self.max_retries):
            file_name = self.client.generate_answer(msg)
            file_name = self._sanitize_string(file_name)
            if self._validate_file_name(file_name):
                return file_name
            else:
                msg.append({"role": "assistant", "content": file_name})
                msg.append(
                    {
                        "role": "user",
                        "content": f"The topic is to long please make it shorter,"
                                   f"max character count: {self.max_file_name_length}.",
                    },
                )

        raise Exception("Max retries exceeded.")
